#change the following according to your needs
from your_simulation_module import your_simulation_function

def simulation_code(kwargs):

    result = your_simulation_function(**kwargs)

    #do stuff with the result
    #if you create an object carrying out the simulation like so:
    #
    #result = sim.your_simulation_function(**kwargs)
    #
    #consider deleting your object afterwards in order to free memory:
    #
    #del sim
    #
    #(be aware that the results from the simulation have to be a deep copy of
    # of data in the simulation object, or else the resultdata will get deleted, too)
    
    return result



from __future__ import print_function
import sys
import qsuite
import os
import ast

if sys.version_info[0] == 2:
    import ConfigParser as cp
elif sys.version_info[0] == 3:
    import configparser as cp
else:
    print("Python version",sys.version_info[0],"not supported")
    sys.exit(1)

def get_qsuite(qsuitefile,init=False):

    if init:
        config = cp.RawConfigParser()
        config.add_section('Files')
        config.set('Files','config','qsuite_config.py')
        config.set('Files','simulation','simulation.py')
        config.set('Files','customwrap',None)
        config.set('Files','additional_files',[''])
        config.set('Files','execute_after_scp',None)
        write_qsuite(config,qsuitefile)
    else:
        config = cp.RawConfigParser()
        config.read(qsuitefile)

    return config

def set_in_qsuite(qsuiteparser,qsuitefile,mode,set_to_thing):

    if mode == 'config' or mode == 'simulation' or mode == 'customwrap':
        opt = mode
    elif mode == 'add':
        opt = "additional_files"
    elif mode == 'exec':
        opt = "execute_after_scp"
    else:
        print("Unkown mode:",mode)
        sys.exit(1)

    if type(set_to_thing) in [list,tuple]:
        stuff_exists = True
        for fname in set_to_thing:
            this_exists = os.path.exists(fname)
            stuff_exists = stuff_exists and this_exists
            if not this_exists:
                print("File", set_to_thing,"doesn't exist.")
        if not stuff_exists:
            sys.exit(1)
    else:
        stuff_exists = os.path.exists(set_to_thing)

    if stuff_exists:
        if mode == "add":
            add_files = ast.literal_eval(qsuiteparser.get('Files','additional_files'))

            if add_files[0] == '':
                add_files.pop(0)    

            add_files.extend(set_to_thing)
            add_files = list(set(add_files))
            set_to_thing = add_files

        qsuiteparser.set('Files',opt,set_to_thing)
        write_qsuite(qsuiteparser,qsuitefile)
    else:
        print("File",set_to_thing,"does not exist")
        sys.exit(1)

def rm_in_qsuite(qsuiteparser,qsuitefile,things_to_rm):
    add_files = ast.literal_eval(qsuiteparser.get('Files','additional_files'))
    exec_file = qsuiteparser.get('Files','execute_after_scp')
    custom_file = qsuiteparser.get('Files','customwrap')

    changed = False
    
    for thing_to_rm in things_to_rm: 
        if thing_to_rm in add_files:
            ndx = add_files.index(thing_to_rm)
            add_files.pop(add_files.index(thing_to_rm))
            if len(add_files)==0:
                add_files.append('')
            qsuiteparser.set('Files','additional_files',add_files)
            changed = True

        if thing_to_rm == exec_file:
            qsuiteparser.set('Files','execute_after_scp',None)
            changed = True

        if thing_to_rm == custom_file:
            qsuiteparser.set('Files','customwrap',None)
            changed = True

    if changed:
        write_qsuite(qsuiteparser,qsuitefile)

def write_qsuite(qsuiteparser,qsuitefile):
    qsuiteparser.write(open(qsuitefile,'w'))



from __future__ import print_function
import paramiko
import select
import os
import sys

def ssh_command(ssh,command,noprint=False):
    """
    this is adapted from code by Sebastian Dahlgren
    http://sebastiandahlgren.se/2012/10/11/using-paramiko-to-send-ssh-commands/
    """

    # Send the command (non-blocking)
    print("ssh> " + command)
    stdin,stdout,stderr = ssh.exec_command(command)

    # Wait for the command to terminate
    last_line = ''
    complete_received = ''
    while not stdout.channel.exit_status_ready():
    # Only print data if there is data to read in the channel
        while stdout.channel.recv_ready():
            rl, wl, xl = select.select([stdout.channel], [], [], 0.0)
            if len(rl) > 0:
        # Print data from stdout
                recv = stdout.channel.recv(1024).decode('ascii')
                lines = recv.split('\n')
                if len(lines)>1:
                    first_lines = '\n'.join(lines[:-1]) + '\n'
                    received = last_line + first_lines
                    last_line = lines[-1]
                else:
                    last_line = ''
                    received = recv
                complete_received += received
                if not noprint:
                    sys.stdout.write(received)

    while stdout.channel.recv_ready():
        rl, wl, xl = select.select([stdout.channel], [], [], 0.0)
        if len(rl) > 0:
            # Print data from stdout
            recv = stdout.channel.recv(1024).decode('ascii')
            lines = recv.split('\n')
            if len(lines)>1:
                first_lines = '\n'.join(lines[:-1]) + '\n'
                received = last_line + first_lines
                last_line = lines[-1]
            else:
                last_line = ''
                received = recv
            complete_received += received
            if not noprint:
                sys.stdout.write(received)

                
    if (last_line is not None) and (last_line != "") and (not last_line.endswith("\n")):
        if not noprint:
            sys.stdout.write(last_line+"\n")
        complete_received += last_line+"\n"

    err = '\n'.join(stderr.read().decode('utf-8').split('\n')[:-1])
    if err != '' and not noprint:
        print(err)

    return complete_received


def ssh_connect(cf):
    """
    this is adapted from code by Sebastian Dahlgren
    http://sebastiandahlgren.se/2012/10/11/using-paramiko-to-send-ssh-commands/
    """
    try:
        ssh = paramiko.SSHClient()
        ssh.load_system_host_keys()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(cf.server,username=cf.username)
        print("Connected to %s" % cf.server)
    except paramiko.AuthenticationException as e:
        print("Authentication failed when connecting to %s" % cf.server)
        print("error:",e)
        sys.exit(1)
    except Exception as e:
        print("Couldn't establish an ssh connection to %s" % cf.server)
        print("error:", e)
        sys.exit(1)

    return ssh

def print_progress(transferred, toBeTransferred):
    progress = transferred / float(toBeTransferred)
    _update_progress(progress)

def _update_progress(progress, bar_length=40, status=""):        
    """
    This function was coded by Marc Wiedermann (https://github.com/marcwie)

    Draw a progressbar according to the actual progress of a calculation.

    Call this function again to update the progressbar.

    :type progress: number (float)
    :arg  progress: Percentile of progress of a current calculation.

    :type bar_length: number (int)
    :arg  bar_length: The length of the bar in the commandline.

    :type status: str
    :arg  status: A message to print behing the progressbar.
    """
    block = int(round(bar_length*progress))
    text = "\r[{0}] {1:4.1f}% {2}".format("="*block + " "*(bar_length-block),
                                     round(progress, 3)*100, status)
    sys.stdout.write(text)
    sys.stdout.flush()
    #if progress >= 1:
    #    sys.stdout.write("\n")

def sftp_put_files(ssh,cf,files_destinations):

    #ssh_command(ssh,
    #            "mkdir -p "+cf.serverpath+"; "+\
    #            "mkdir -p "+cf.resultpath+"; "+\
    #            "mkdir -p "+cf.serverpath+"/output")
    
    #additional_directories = set([cf.serverpath, cf.resultpath, cf.serverpath+"/output"])
    all_directories = set()
    mkdir_p_string = "mkdir -p "+cf.serverpath+"/output; mkdir -p "+cf.resultpath+"; "

    for f,d in files_destinations:
        directory = '/'.join(d.split('/')[:-1])

        if directory not in all_directories:
            all_directories.add(directory)
            mkdir_p_string += 'mkdir -p ' + directory + "; "

    ssh_command(ssh,mkdir_p_string)
    ftp = ssh.open_sftp()

    for f,d in files_destinations:
        print(" "+f+"\n =>"+d)
        ftp.put(f,d,callback=print_progress)
        sys.stdout.write("\n")

def sftp_get_files(ssh,cf,files_destinations):

    ftp = ssh.open_sftp()

    for f,d in files_destinations:
        print(" "+f+"\n =>"+d)
        ftp.get(f,d,callback=print_progress)
        sys.stdout.write("\n")



import os 

#=========== SIMULATION DETAILS ========
projectname = "project"
basename = "experimentname"

seed = -1
N_measurements = 1

measurements = range(N_measurements)
params1 = range(3)
params2 = range(3)
params3 = range(3)
params4 = range(3)
params5 = range(3)
params6 = range(3)

external_parameters = [
                        ( 'p1', params1[:2]   ),
                        ( 'p2', params2       ),
                        ( None   , measurements ),
                      ]
internal_parameters = [
                        ('p3', params3[:1]),
                        ('p3', params4[:]),
                      ]
standard_parameters = [
                        ( 'p5', params5[1] ),
                        ( 'p6', params6[2] ),
                      ]

only_save_times = False

#============== QUEUE ==================
queue = "SGE"
memory = "1G"
priority = 0

#============ CLUSTER SETTINGS ============
username = "user"
server = "localhost"
useratserver = username + u'@' + server

shell = "/bin/bash"
pythonpath = "python"
name = basename + "_NMEAS_" + str(N_measurements) + "_ONLYSAVETIME_" + str(only_save_times)
serverpath = "/home/"+username +"/"+ projectname + "/" + name 
resultpath = serverpath + "/results"

#============== LOCAL SETTINGS ============
localpath = os.path.join(os.getcwd(),"results_"+name)
n_local_cpus = 1

#========================
git_repos = [
                ( "/path/to/repo", pythonpath + " setup.py install --user" )
            ]




"""
A collection of useful tools.
"""

from qsuite import qconfig
import numpy as np

def change_result_parameter_order(results, new_order, old_order=None, ignore_measurement_axis=False):
    """
    Take the result array and move around the axes such that it corresponds to 
    the order of parameters in `new_order` (a list of the parameter names as
    defined in `external_parameters` and `internal_parameters`).
    """

    res = results
    is_list = type(results) is list
    if is_list:
        res = np.array(res)

    if old_order is None:
        cf = qconfig()
        params = [ _[0] for _ in cf.external_parameters + cf.internal_parameters ]
    else:
        params = old_order

    if ignore_measurement_axis and None in params:
        params.pop(params.index(None))
    
    new_indices = np.array([new_order.index(p) for p in params])
    if not np.all(np.sort(new_indices) == np.arange(len(params))):
        raise ValueError("Every element in the list `new order` needs to have a corresponding element in the parameter configuration.")

    res = np.moveaxis(res, np.arange(len(params)), new_indices)

    if is_list:
        res = res.tolist()

    return res



def change_meanerr_parameter_order(meanerr, new_order, old_order=None):
    return change_result_parameter_order(meanerr, new_order, old_order=old_order, ignore_measurement_axis=True)



import os

__version__ = "0.4.14"

__author__ = "Benjamin F. Maier"
__copyright__ = "Copyright 2015-2021, " + __author__
__credits__ = [__author__]
__license__ = "MIT"
__maintainer__ = __author__
__email__ = "contact@benmaier.org"
__status__ = "Development"

customdir = os.path.join(os.path.expanduser("~"),".qsuite")

def rm(path):
    if os.path.exists(path):
        os.remove(path)

def mkdirp(path):
    if not os.path.exists(path):
        os.mkdir(path)

from .qconfig import *
from .template import *
from .qsuiteparse import *
from .ssh import *
from .submitjob import *
from .printparams import *
import qsuite.queuesys as queuesys



from __future__ import print_function
from numpy import *
import os
import sys
#import gzip

try:
    # Python 2
    import cPickle as pickle
except ImportError:
    # Python 3
    import pickle

class _progress_updater():

    def __init__(self):
        self.text = ""

    def update_progress(self,progress, bar_length=40, status=""):        
        """
        This function was coded by Marc Wiedermann (https://github.com/marcwie)

        Draw a progressbar according to the actual progress of a calculation.

        Call this function again to update the progressbar.

        :type progress: number (float)
        :arg  progress: Percentile of progress of a current calculation.

        :type bar_length: number (int)
        :arg  bar_length: The length of the bar in the commandline.

        :type status: str
        :arg  status: A message to print behing the progressbar.
        """
        block = int(round(bar_length*progress))
        text = "\r[{0}] {1:4.1f}% {2}".format("="*block + " "*(bar_length-block),
                                         round(progress, 3)*100, status)
        if text != self.text:
            self.text = text
            sys.stdout.write(text)
            sys.stdout.flush()
            if progress >= 1:
                sys.stdout.write("\n")

def _get_timeleft_string(t):
    d, remainder = divmod(t,24*60*60)
    h, remainder = divmod(remainder,60*60)
    m, s = divmod(remainder,60)
    t = [d,h,m,s]
    t_str = ["%dd", "%dh", "%dm", "%ds"]
    it = 0 
    while it<4 and t[it]==0.:
        it += 1
    text = (" ".join(t_str[it:it+2])) % tuple(t[it:it+2])
    return text

def wrap_results(is_local=False,localrespath="current_results"):

    if not is_local:
        from qconfig import qconfig
    else:
        from qsuite import qconfig

    cf = qconfig()

    if not is_local:
        resultpath = cf.resultpath
        outputpath = resultpath
    else:
        resultpath = os.path.join(os.getcwd(),localrespath)
        outputpath = cf.localpath

    #get dimensions of the list space
    pdims = [ len(p[1]) for p in cf.external_parameters ]
    idims = [ len(p[1]) for p in cf.internal_parameters ]

    #prepare result arrays
    results = empty(pdims+idims,dtype=object).flatten()
    times = empty(pdims+idims,dtype=object).flatten()

    #wrap results of the simulations
    loading_successful = True
    missing_array_ids = []
    updater = _progress_updater()
    for j in range(cf.jmax+1):
        #get coords in list space from linear index
        if len(pdims) > 0:
            pcoords = [ [c] for c in list(unravel_index(j, pdims))]
        else:
            pcoords = []

        if os.path.exists(resultpath+"/times_%d.p" % j) and os.path.exists(resultpath+"/results_%d.p" % j):

            if loading_successful:

                with open(resultpath+"/times_%d.p" % j,'rb') as this_file:
                    time = pickle.load(this_file)

                if not cf.only_save_times:
                    with open(resultpath+"/results_%d.p" % j,'rb') as this_file:
                        res = pickle.load(this_file)
        else:
            if loading_successful:
                print("*** Caught Exception: Files missing! Jobs with the following ARRAY IDs did not produce results:")
            missing_array_ids.append(j+1)
            #print("*** Caught exception: %s,%s" % (e.__class__,e))
            loading_successful = False

                
        if loading_successful:
            for i in range(len(time)):
                if len(idims) > 0:
                    icoords = [ [c] for c in list(unravel_index(i, idims)) ]
                else:
                    icoords = []
                flat_index = ravel_multi_index(pcoords+icoords,pdims+idims)[0]
                times[flat_index] = time[i]
                if not cf.only_save_times:
                    results[flat_index] = res[i]

        updater.update_progress((j+1.)/(cf.jmax+1.))

    if loading_successful:
        # convert to lists for better portability
        times = times.reshape(pdims+idims).tolist()
        results = results.reshape(pdims+idims).tolist()

        try:
            #save the wrapped data
            with open(outputpath+"/times.p",'wb') as dumpfile:
                pickle.dump(times,dumpfile)
            if not cf.only_save_times:
                with open(outputpath+"/results.p",'wb') as dumpfile:
                    pickle.dump(results,dumpfile)
        except Exception as e:
            print("couldn't write files")
            print("Error:",e)
            sys.exit(1)

            
        # delete the original files    
        for j in range(cf.jmax+1):
            os.remove(resultpath+"/times_%d.p" % j)
            if not cf.only_save_times:
                os.remove(resultpath+"/results_%d.p" % j)
    else: 
        i = 0
        missing_array_id_strings = []

        while i < len(missing_array_ids):
            this_id = missing_array_ids[i]
            this_str = str(this_id)
            while i+1 < len(missing_array_ids) and missing_array_ids[i+1] == this_id+1:
                i += 1
                this_id = missing_array_ids[i]

            if (i+1 == len(missing_array_ids) or this_id+1 != missing_array_ids[i+1]) and this_str != str(this_id):
                this_str += '-' + str(this_id)

            missing_array_id_strings.append(this_str)
            i += 1

        print(' '.join(missing_array_id_strings))
        sys.exit(1)

if __name__ == "__main__":
    wrap_results()



from __future__ import print_function
import os
import sys
import qsuite
import subprocess
import math 

from functools import reduce
import operator

from qsuite import qconfig
import ast
from qsuite import get_template_file
from qsuite import copy_template
from qsuite import set_default_file 
from qsuite import get_qsuite
from qsuite import set_in_qsuite
from qsuite import rm_in_qsuite
from qsuite import write_qsuite
from qsuite import rm 
from qsuite import reset 
from qsuite import ssh_command 
from qsuite import ssh_connect
from qsuite import make_job_ready
from qsuite import start_job
from qsuite import sftp_put_files
from qsuite import sftp_get_files
from qsuite import mkdirp
from qsuite import print_params
from qsuite import print_status
from qsuite import print_params_and_status
from qsuite.queuesys.job import job
from qsuite.printparams import _get_progress_sftp
import shutil
import re

from operator import itemgetter
from itertools import groupby

from pathos.multiprocessing import ProcessingPool as Pool
from functools import partial

try:
    # Python 2
    import cPickle as pickle
except ImportError:
    # Python 3
    import pickle
    basestring = str

if sys.version_info[0] == 3:
    raw_input = input

def update_git(cf,ssh):
    repos = cf.git_repos

    #ssh.connect(cf.server,username=cf.username)
    repostring = ''
    for repo in cf.git_repos:

        if len(repo)==3:
            repostring += 'if [ ! -d "%s" ]; then\ngit clone %s %s\nfi\n' % (repo[0],repo[2],repo[0])

        repostring += "cd %s; git fetch; git pull; %s; " % tuple(repo[:2])

    if len(cf.git_repos)>0 :
        ssh_command(ssh,repostring)


def wrap_results(cf,ssh):
    ssh_command(ssh, "cd " + cf.serverpath + "; " + cf.pythonpath + " wrap_results.py;")
    ssh_command(ssh, "cd " + cf.serverpath + "/results; gzip results.p" )
    ssh_command(ssh, "cd " + cf.serverpath + "/results; gzip times.p" )
    custom_wrap_results(cf,ssh)

def custom_wrap_results(cf,ssh):
    if "custom_wrap_results.py" in cf.files_to_scp:
        ssh_command(ssh, "cd " + cf.serverpath + "; " + cf.pythonpath + " custom_wrap_results.py;")

def wrap_local(cf):

    files = list(cf.files_to_scp.values()) + cf.additional_files_to_scp
    mkdirp(cf.localpath)
    for f in files:
        shutil.copy2(f,cf.localpath)

def qstat(cf,ssh,args):
    if len(args)==0:
        flag = "-u " + cf.username
        ssh_command(ssh,"qstat "+flag)

    elif args[0] == "job":
        if cf.queue=="SGE":
            flag = "-j"
        elif cf.queue=="PBS":
            flag = ""
        else:
            print("Queue",cf.queue,"not supported.")
            sys.exit(1)
        ssh_command(ssh,"qstat "+flag+" $(cat "+cf.serverpath+"/.jobid)")

    elif args[0] == "all":
        if cf.queue=="SGE":
            flag = '-u "*"'
        elif cf.queue=="PBS":
            flag = "-Q"
        else:
            print("Queue",cf.queue,"not supported.")
            sys.exit(1)
        ssh_command(ssh,"qstat "+flag)


def init(qsuitefile,opts):
    qsuiteparser = get_qsuite(qsuitefile,init=True)
    copy_template("config",opts)
    copy_template("simulation",opts)
    write_qsuite(qsuiteparser,qsuitefile)



def main():

    #create the custom dir if it doesn't exist yet

    if len(sys.argv)==1:
        print("Yes, I'm here! Give me a command!")
        sys.exit(1)

    args = [ a for a in sys.argv[1:] if not a.startswith("-") ]
    opts = [ a for a in sys.argv[1:] if a.startswith("-") ]

    cmd = args[0]
    cwd = os.getcwd()
    qsuitefile = os.path.join(cwd,".qsuite")

    cf = None
    qsuiteparser = None

    git_cmds = ["git","gitupdate","git_update","updategit","update_git"]
    prep_cmds = [ "prepare", "make", "makeready" ]
    submit_cmds = [ "submit", "start", "submit_job"]
    set_cmds = ["set"]
    reset_cmds = ["reset", "resetdefault"]
    add_cmds = ["add"]
    rm_cmds = ["rm","remove"]
    wrap_cmds = ["wrap","wrapresults","wrap_results"]
    qstatus_cmds = ["qstat","qstatus"]
    status_cmds = ["stat","status"]
    ssh_cmds = ["ssh"]
    sftp_cmds = ["sftp","scp","ftp","put"]
    customwrap_cmds = ["customwrap"]
    get_cmds = ["get"]
    test_cmds = ["test", "local"]
    err_cmds = ["err","error"]
    param_cmds = ["params"]
    convert_cmds = ["convert"]
    estimation_cmds = ["estimate", "estimatespace", "data"]

    if cmd in ["init","initialize"]:
        if len(args)==1:
            if not os.path.exists(qsuitefile):
                init(qsuitefile,opts)
                sys.exit(0)
            else:
                yn = raw_input("There's already a .qsuite file carrying configurations. Do you want to override it (y/n)? ")
                yn = yn.lower()[0]
                if yn == "y":
                    init(qsuitefile,opts)
                sys.exit(0)
        elif len(args)>1:
            if args[1] in ["customwrap", "customwrapper"]:
                copy_template("customwrap",opts)
                sys.exit(0)
            else:
                print("Unknown option",args[1])
                sys.exit(1)
    elif not os.path.exists(os.path.join(cwd,".qsuite")) and not cmd in convert_cmds:
        if (    (cmd not in set_cmds) \
             or (len(args)<2) \
             or (not args[1].startswith("default")) ):

            #if there's no ".qsuite" file yet, stop operating
            print("Not initialized yet!")
            sys.exit(1)
    elif cmd in convert_cmds:
        #convert to numpy and calculate mean and error
        #only works in results directory!
        cf = qconfig()

        cmds = args[1:]
        if len(cmds)==0:
            cmds = ["numpy"]

        # numpy is imported here because I don't want to make qsuite
        # dependent on it, and this will give an error on runtime,
        # not when qsuite is started
        import numpy as np

        data_already_loaded = False

        if "numpy" in cmds or \
           (any([c in cmds for c in ["meanerr","nanmeanerr","meanstd","nanmeanstd","defaultpercentiles","percentiles","quartiles"]]) and not os.path.exists("results.npy")):

            if not os.path.exists(os.path.join(cwd, "results.p")):
                p = subprocess.Popen(["gzip", "-d", "results.p.gz"])
                exit_code = p.wait() 
                if exit_code == 1:
                    print("error while using gzip -d results.p.gz")
                    sys.exit(1)

            try:
                data = pickle.load(open("results.p","rb"))
            except UnicodeDecodeError:
                data = pickle.load(open("results.p","rb"), encoding='latin1')

            data = np.array(data)
            np.save("results.npy",data)
            data_already_loaded = True

        if "meanerr" in cmds:
            if not data_already_loaded:
                data = np.load("results.npy")

            axis = None
            if isinstance(axis,basestring) or axis is None:
                axis = (cf.parameter_names+cf.internal_names).index(axis)

            N = len((cf.external_parameters+cf.internal_parameters)[axis][1])
            mn, err = data.mean(axis=axis), data.std(axis=axis)/np.sqrt(N)
            np.savez("./results_mean_err.npz",mean=mn,err=err)

        if "nanmeanerr" in cmds:
            if not data_already_loaded:
                data = np.load("results.npy")

            axis = None
            if isinstance(axis,basestring) or axis is None:
                axis = (cf.parameter_names+cf.internal_names).index(axis)

            N = len((cf.external_parameters+cf.internal_parameters)[axis][1])
            mn, err = np.nanmean(data,axis=axis), np.nanstd(data,axis=axis)/np.sqrt(N)
            np.savez("./results_mean_err.npz",mean=mn,err=err)

        if "meanstd" in cmds:
            if not data_already_loaded:
                data = np.load("results.npy")

            axis = None
            if isinstance(axis,basestring) or axis is None:
                axis = (cf.parameter_names+cf.internal_names).index(axis)

            mn, err = data.mean(axis=axis), data.std(axis=axis)
            np.savez("./results_mean_std.npz",mean=mn,std=err)

        if "nanmeanstd" in cmds:
            if not data_already_loaded:
                data = np.load("results.npy")

            axis = None
            if isinstance(axis,basestring) or axis is None:
                axis = (cf.parameter_names+cf.internal_names).index(axis)

            N = len((cf.external_parameters+cf.internal_parameters)[axis][1])
            mn, err = np.nanmean(data,axis=axis), np.nanstd(data,axis=axis)
            np.savez("./results_mean_std.npz",mean=mn,std=err)

        percentiles_name = None

        if "defaultpercentiles" in cmds:
            percentiles_name = "defaultpercentiles"
            cmds = ["percentiles"]
            cmds.extend(["2.5","16","50","84","97.5"])

        if "quartiles" in cmds:
            percentiles_name = "quartiles"
            cmds = ["percentiles"]
            cmds.extend(["25","50","75"])

        if "percentiles" in cmds:

            if percentiles_name is None:
                percentiles_name = "percentiles"
            ndx = cmds.index("percentiles")
            percentiles = cmds[ndx+1:]
            if len(percentiles) == 0:
                print("No percentiles provided. Please provide percentiles as list or use `defaultpercentiles` or `quartiles`")
                sys.exit(1)
            fperc = [float(p) for p in percentiles]

            if not data_already_loaded:
                data = np.load("results.npy")

            axis = None
            if isinstance(axis,basestring) or axis is None:
                axis = (cf.parameter_names+cf.internal_names).index(axis)

            data_percentiles = np.percentile(data, fperc, axis=axis)
            results = {}

            for i, perc in enumerate(percentiles):
                results[perc] = data_percentiles[i]
            np.savez('./results_'+percentiles_name+".npz",**results)



        if not os.path.exists(os.path.join(cwd, "results.p.gz"))\
            and os.path.exists(os.path.join(cwd, "results.p")):
            p = subprocess.Popen(["gzip", "results.p"])
            exit_code = p.wait()
            if exit_code == 1:
                print("error while using gzip results.p")
                sys.exit(1)

    elif cmd in (git_cmds + submit_cmds + prep_cmds + reset_cmds + add_cmds + rm_cmds +\
                 set_cmds + wrap_cmds + status_cmds + ssh_cmds + sftp_cmds + customwrap_cmds +\
                 get_cmds + test_cmds + param_cmds + qstatus_cmds + err_cmds + estimation_cmds):

        #I do this to prevent that the default stuff can only be set in an initialized dir
        if not (cmd in set_cmds and len(args)>1 and args[1].startswith("default")):
            qsuiteparser = get_qsuite(qsuitefile)

        if cmd in set_cmds:
            if len(args)>2:

                thing_to_set = args[1]
                set_to_thing = args[2]

                if thing_to_set in ["cfg", "config", "configuration"]:
                    set_in_qsuite(qsuiteparser,qsuitefile,"config",set_to_thing)
                elif thing_to_set in ["sim", "simulation"]:
                    set_in_qsuite(qsuiteparser,qsuitefile,"simulation",set_to_thing)
                elif thing_to_set in ["customwrap", "customwrapper"]:
                    set_in_qsuite(qsuiteparser,qsuitefile,"customwrap",set_to_thing)
                elif thing_to_set in ["exec", "execute", "exe"]:
                    set_in_qsuite(qsuiteparser,qsuitefile,"exec",set_to_thing)

                elif thing_to_set.startswith("default"):
                    if len(args)>2:

                        file_to_set = args[2]

                        if thing_to_set in ["defaultcfg", "defaultconfig", "defaultconfiguration"]:
                            set_default_file("config",file_to_set)
                        elif thing_to_set in ["defaultsim", "defaultsimulation"]:
                            set_default_file("simulation",file_to_set)
                        elif thing_to_set in ["defaultwrap", "defaultcustomwrap", "defaultcustomwrapper"]:
                            set_default_file("customwrap",file_to_set)
                        sys.exit(0)
                    else:
                        print("No default file given.")
                        sys.exit(1)
                else:
                    print("Option "+ thing_to_set +" not known.")
                    sys.exit(1)
                sys.exit(0)
            elif len(args)==2:
                thing_to_set = args[1]
                if thing_to_set in ["custom_wrap_results.py"]:
                    set_in_qsuite(qsuiteparser,qsuitefile,"customwrap",thing_to_set)
            else:
                print("Nothing to set!")
                sys.exit(1)

        elif cmd in reset_cmds:
            if len(args)>1:

                things_to_reset = args[1:]
                for thing_to_reset in things_to_reset:
                    if thing_to_reset in ["defaultcfg", "defaultconfig", "defaultconfiguration"]:
                        reset("config")
                    elif thing_to_reset in ["defaultsim", "defaultsimulation"]:
                        reset("simulation")
                    elif thing_to_reset in ["defaultwrap", "defaultcustomwrap", "defaultcustomwrapper"]:
                        reset("customwrap")
                sys.exit(0)
            else:
                print("Nothing to reset given.")
                sys.exit(1)

        elif cmd in add_cmds:

            if len(args)>1:
                thing_to_set = args[1:]

                if "custom_wrap_results.py" in thing_to_set:
                    set_in_qsuite(qsuiteparser,qsuitefile,"customwrap","custom_wrap_results.py")
                    thing_to_set.pop(thing_to_set.index("custom_wrap_results.py"))

                if len(thing_to_set)>0:
                    set_in_qsuite(qsuiteparser,qsuitefile,"add",thing_to_set)

                sys.exit(0)
            else:
                print("Nothing to add!")
                sys.exit(1)

        elif cmd in rm_cmds:
            if len(args)>1:
                thing_to_set = args[1:]
                rm_in_qsuite(qsuiteparser,qsuitefile,thing_to_set)
                sys.exit(0)
            else:
                print("Nothing to remove!")
                sys.exit(1)

        elif cmd in test_cmds:

            cf = qconfig(qsuiteparser=qsuiteparser)
            print(type(cf))

            if cmd == "test":

                if len(args)==1:
                    jobid = 0
                    respath = os.path.join(os.getcwd(),".test")
                elif len(args)==2:
                    jobid = int(args[1])
                    respath = os.path.join(os.getcwd(),".test")
                elif len(args)>2:
                    jobid = int(args[1])
                    respath = args[2]

                sys.path.insert(1,os.getcwd()) #add current working directory to import paths
                job(jobid,respath,cf)

            elif cmd == "local":

                N_jobs = cf.get_number_of_jobs()
                jobids = range(N_jobs)
                respath = "current_results"

                def local_job(j_id,N_jobs=None,respath=None,cf=None):
                    print(j_id+1,"/",N_jobs)
                    job(j_id,respath,cf)

                wrap_local(cf)

                sys.path.insert(1,os.getcwd()) #add current working directory to import paths

                if hasattr(cf,"n_local_cpus"):
                    ncpu = cf.n_local_cpus
                else:
                    ncpu = 1

                partial_job = partial(local_job,N_jobs=N_jobs,respath=respath,cf=cf)
                print("Starting local job with", ncpu, "CPUs")
                pool = Pool(ncpu)    
                pool.map(partial_job,jobids)

                from qsuite.queuesys.wrap_results import wrap_results as _wrap
                _wrap(is_local=True,localrespath=respath)

            sys.exit(0)     

        elif cmd in param_cmds:
            cf = qconfig(qsuiteparser=qsuiteparser)
            print_params(cf)

        elif cmd in estimation_cmds:

            if len(args)>1:
                bytestring = args[1]
            else:
                print("No number of bytes per parameter combination given.")
                sys.exit(1)

            cf = qconfig(qsuiteparser=qsuiteparser)

            #get dimensions of the list space
            pdims = [ len(p[1]) for p in cf.external_parameters ]
            idims = [ len(p[1]) for p in cf.internal_parameters ]

            def get_size_string(number_of_bytes):
                if number_of_bytes == 0:
                    return "0B"
                names = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
                i = int(math.floor(math.log(number_of_bytes, 1024)))
                p = math.pow(1024, i)
                s = round(number_of_bytes / p, 2)
                return "%s %s" % (s, names[i])

            number_of_bytes_per_combination = eval(bytestring)
            result_internal = reduce(operator.mul, idims, 1) * number_of_bytes_per_combination
            number_of_external = reduce(operator.mul, pdims, 1)
            result_external = number_of_external * result_internal

            print("This job will generate {0:d} jobs, each of an estimated size of {1}.".format(number_of_external, get_size_string(result_internal)))
            print("This means that the total size of the generated data will be an estimated {0}".format(get_size_string(result_external)))


        elif cmd in git_cmds + prep_cmds + submit_cmds + wrap_cmds + status_cmds +\
                    ssh_cmds + sftp_cmds + customwrap_cmds + get_cmds + qstatus_cmds + err_cmds:

            cf = qconfig(qsuiteparser=qsuiteparser)
            ssh = ssh_connect(cf)

            # replace directory keyword with path
            for iarg, arg in enumerate(args):
                args[iarg] = args[iarg].replace("DIR",cf.serverpath)
         
            if cmd in git_cmds:
                update_git(cf,ssh)
                sys.exit(0)        
            elif cmd in prep_cmds:
                update_git(cf,ssh)
                make_job_ready(cf,ssh)
                wrap_local(cf)
                sys.exit(0)
            elif cmd in submit_cmds:
                cmds = args[1:]
                if len(cmds)>0:
                    try:
                        array_id = [ int(c) if ("-" not in c) else [ int(rangeid) for rangeid in c.split("-") ] for c in cmds]
                        print("Using array IDs "+str(array_id)+". Beware! Array IDs start counting at 1.")
                    except ValueError:
                        if any([ "err" in c or "wait" in c for c in cmds ]):

                            key_words = []
                            if any([ "err" in c for c in cmds ]):
                                key_words.append("ERR")
                            if any([ "wait" in c for c in cmds ]):
                                key_words.append("WAIT")

                            print("Fetching array IDs that produced errors with keywords", " and ".join(key_words))
                            #print(key_words)
                            progresses = _get_progress_sftp(cf,ssh)
                            array_id = []
                            last_id = 0
                            for j in range(len(cf.parameter_list)):
                                print(j,progresses[j][0].upper())
                                if any( kw in progresses[j][-1].upper() or kw in progresses[j][0].upper() for kw in key_words ):                                    
                                    array_id.append(j+1)

                            print(array_id)
                            if len(array_id)>0:
                                ranges = []
                                for k, g in groupby( enumerate(array_id), lambda IX:IX[0]-IX[1]):
                                    group = list(map(itemgetter(1), g))
                                    if group[0]==group[-1]:
                                        ranges.append(group[0])
                                    else:
                                        ranges.append((group[0], group[-1]))

                                print("Found erros in jobs with Array IDs", ranges)
                                array_id = ranges
                            else:
                                print("No erroneous IDs found.")
                                sys.exit(1)

                        else:
                            print("Unknown submission command",cmds[0],". Aborting.")
                            sys.exit(1)
                else:
                    array_id = None
                    update_git(cf,ssh)

                make_job_ready(cf,ssh,array_id)
                
                if array_id is None:
                    wrap_local(cf)

                start_job(cf,ssh,array_id)

                sys.exit(0)
            elif cmd in wrap_cmds:
                if len(args) == 1:
                    wrap_results(cf,ssh)
                    ssh_command
                elif args[1] == "local":
                    respath = 'current_results'
                    from qsuite.queuesys.wrap_results import wrap_results as _wrap
                    _wrap(is_local=True,localrespath=respath)

                sys.exit(0)
            elif cmd in customwrap_cmds:
                custom_wrap_results(cf,ssh)
                sys.exit(0)
            elif cmd in qstatus_cmds:
                cmds = args[1:]
                qstat(cf,ssh,cmds)
                sys.exit(0)
            elif cmd in status_cmds:
                cmds = args[1:]
                if len(cmds)>0:
                    print_params_and_status(cf,ssh)
                else:
                    print_status(cf,ssh)
                sys.exit(0)
            elif cmd in err_cmds:
                cmds = args[1:]
                if len(cmds)>0:
                    array_id = str(int(cmds[0]))
                else:
                    array_id = "1"
                print("Using array ID "+array_id+". Beware! Array IDs start counting at 1.")
                ssh_command(ssh,"cat "+cf.serverpath+"/output/"+cf.basename+".sh.e$(cat "+cf.serverpath+"/.jobid)."+array_id)
                sys.exit(0)
            elif cmd in ssh_cmds:
                cmds = args[1:]
                ssh_command(ssh," ".join(cmds))
                sys.exit(0)
            elif cmd in sftp_cmds:
                files = args[1:]
                files_dests = [ (f,cf.serverpath+"/"+f) for f in files ]
                sftp_put_files(ssh,cf,files_dests)
            elif cmd in get_cmds:
                #get the wrapped result
                if len(args)>1 and args[1] not in ["all","results","allresults"]:
                    filenames = args[1:]
                    local_filenames = [ f.split('/')[-1] for f in filenames ]
                    files_dests = [ (cf.serverpath+"/"+f, os.path.join(cwd,cf.localpath,local_f)) \
                                    for f,local_f in zip(filenames,local_filenames) ]
                    sftp_get_files(ssh,cf,files_dests)
                else:
                    get_all =  len(args)>1 and args[1] in ["all","allresults"]

                    pattern_res = re.compile(r'result.*\.p*$')
                    pattern_tim = re.compile(r'time.*\.p*$')
                    resultstring = ssh_command(ssh,"ls "+cf.resultpath)
                    resultstring = resultstring.replace("\n"," ")
                    resultlist = [ f for f in resultstring.split(" ") if (f!="" and not f.startswith(".")) ]
                    resultlist = [ r for r in resultlist if not ((not get_all) and ((pattern_res.match(r)) or (pattern_tim.match(r)))) ]

                    files_dests = [ (cf.resultpath+"/"+f, os.path.join(cwd,cf.localpath,f)) for f in resultlist ]
                    sftp_get_files(ssh,cf,files_dests)

            else:
                pass
    else:
        print("Command",cmd,"unknown.")





from __future__ import print_function
import shutil
import os
import sys
import qsuite
from qsuite import rm

if sys.version_info[0] == 2:
    pass
elif sys.version_info[0] == 3:
    raw_input = input
else:
    print("Python version",sys.version_info[0],"not supported")
    sys.exit(1)

def mkdirp_customdir():
    if not os.path.exists(qsuite.customdir):
        os.makedirs(qsuite.customdir)

def get_template_file(filename):

    #check if there exists a custom template
    potential_file = os.path.join(qsuite.customdir,filename)
    if os.path.exists(potential_file):
        sourcefile = potential_file
    else:
        sourcefile = os.path.join(qsuite.__path__[0],"templates",filename)

    return sourcefile

def reset(mode):

    if mode == "config":
        filename = "qsuite_config.py"
    elif mode == "simulation":
        filename = "simulation.py"
    elif mode == "customwrap":
        filename = "custom_wrap_results.py"
    else:
        print("copy_template(): no such mode as", mode)
        sys.exit(1)

    targetfile = os.path.join(qsuite.customdir,filename)

    if os.path.exists(targetfile):
        yn = raw_input("This action deletes the current default "+mode+" file. Do you want to proceed (y/n)? ")
        yn = yn.lower()[0]
        delete_file = (yn == "y")
        if delete_file:
            qsuite.rm(targetfile)


def copy_template(mode,options=[]):

    if mode == "config":
        filename = "qsuite_config.py"
    elif mode == "simulation":
        filename = "simulation.py"
    elif mode == "customwrap":
        filename = "custom_wrap_results.py"
    else:
        print("copy_template(): no such mode as", mode)
        sys.exit(1)

    sourcefile = get_template_file(filename)

    #target file is the current working directory
    targetfile = os.path.join(os.getcwd(),filename)

    #copy template to cwd
    if not os.path.exists(targetfile) or "-f" in options:
        shutil.copy2(sourcefile,targetfile)

        if targetfile.endswith(".py"):
            rm(targetfile+"c") #remove class files

        print("Initialized",mode,"as",filename)
    else:
        print("The file",filename,"already exists. Remove it first or use the '-f' flag to force copying the template.")




def set_default_file(mode,sourcefile):

    if mode == "config":
        filename = "qsuite_config.py"
    elif mode == "simulation":
        filename = "simulation.py"
    elif mode == "customwrap":
        filename = "custom_wrap_results.py"
    else:
        print("copy_template(): no such mode as", mode)
        sys.exit(1)

    mkdirp_customdir()

    potential_file = os.path.join(qsuite.customdir,filename)
    copy_file = True

    if os.path.exists(potential_file):
        yn = raw_input("There's already a custom " + mode + " file set. Do you want to override it (y/n)? ")
        yn = yn.lower()[0]
        copy_file = (yn == "y")

    if not os.path.exists(sourcefile):
        sourcefile = os.path.join(os.getcwd(),sourcefile)

    if copy_file:
        shutil.copy2(sourcefile,potential_file)



from __future__ import print_function
import sys
import time
import os
import traceback

from numpy import mean
from numpy import std
from numpy import sqrt

try:
    # Python 2
    from itertools import izip
    import cPickle as pickle
except ImportError:
    # Python 3
    import pickle
    izip = zip

if os.path.exists("./qconfig.py"):
    from qconfig import qconfig
else:
    from qsuite import qconfig

def _get_progress_fn(cf, j):
    str_len = str(len(str(len(cf.parameter_list)-1)))
    fmt = "%0"+str_len+"d_%d"
    return cf.serverpath + ("/output/progress_"+fmt) % (j, len(cf.parameter_list)-1)

def _update_progress(progress, bar_length=40, status=""):        
        """
        This function was coded by Marc Wiedermann (https://github.com/marcwie)

        Draw a progressbar according to the actual progress of a calculation.

        Call this function again to update the progressbar.

        :type progress: number (float)
        :arg  progress: Percentile of progress of a current calculation.

        :type bar_length: number (int)
        :arg  bar_length: The length of the bar in the commandline.

        :type status: str
        :arg  status: A message to print behing the progressbar.
        """
        block = int(round(bar_length*progress))
        text = "\r[{0}] {1:4.1f}% {2}".format("="*block + " "*(bar_length-block),
                                         round(progress, 3)*100, status)
        sys.stdout.write(text)
        sys.stdout.flush()
        if progress >= 1:
            sys.stdout.write("\n")

def _get_timeleft_string(t):
    d, remainder = divmod(t,24*60*60)
    h, remainder = divmod(remainder,60*60)
    m, s = divmod(remainder,60)
    t = [d,h,m,s]
    t_str = ["%dd", "%dh", "%dm", "%ds"]
    it = 0 
    while it<4 and t[it]==0.:
        it += 1
    text = (" ".join(t_str[it:it+2])) % tuple(t[it:it+2])
    return text
    

def _update_progress_file(progress_id, N_id, times, filename, bar_length=40):

    progress = (progress_id+1.) / float(N_id)
    block = int(round(bar_length*progress))
    if progress_id==N_id-1:
        timeleft = "done"
    elif len(times)>0:
        timeleft = int((N_id-progress_id-1) * mean(times))
        timeleft = _get_timeleft_string(timeleft)
    else:
        timeleft = "no estimate yet"
    
    text = "[{0}] {1:4.1f}%__{2}\n".format("="*block + " "*(bar_length-block),
                                     round(progress, 3)*100, timeleft)

    progressfile = open(filename,"w")
    progressfile.write(text)
    progressfile.close()

def job(j,resultpath=None,cf=None):

    is_local = cf is not None

    if not is_local:
        cf = qconfig()

    #get the resultpath from arguments
    #if no resultpath is given, it is assumed that a local simulation takes place
    if is_local:
        cf.resultpath = resultpath
        simcode_path = os.path.join(os.getcwd(),cf.files_to_scp["simulation.py"])
    else:
        simcode_path = os.path.join(os.getcwd(),"simulation.py")

    if not os.path.exists(simcode_path):
        print("No simulation file provided!")


    #import the simulation module
    try:
        if sys.version_info[0] == 2:
            import imp
            simcode = imp.load_source("sim",simcode_path)
        elif sys.version_info >= (3,5):
            import importlib.util
            specifications = importlib.util.spec_from_file_location("sim",simcode_path)
            simcode = importlib.util.module_from_spec(specifications)
            specifications.loader.exec_module(simcode)
        else:
            print("Python version",sys.version_info[0],"not supported.")
            sys.exit(1)
    except Exception as e:
        # in case there's an error, write that into the progress file
        if not is_local:
            with open(_get_progress_fn(cf,j),"w") as progressfile:
                progressfile.write("error: "+e.__class__.__name__+"\n")

        traceback.print_exc(file=sys.stderr)
        sys.exit(1) 

    #get kwargs for the simulation of this jobnumber
    job_kwargs = cf.get_kwargs(cf.parameter_names,cf.parameter_list[j])

    #prepare result lists
    results = [ None for i in range(len(cf.internal_parameter_list)) ]
    times = list(results) #copy

    N_int_param = len(results)

    #loop through the internal args
    for ip,internal_params in enumerate(cf.internal_parameter_list):

        #if this is the first run, initiate the progress bar
        if not is_local and ip==0:
            _update_progress_file(ip-1,N_int_param,[],_get_progress_fn(cf,j))

        #wrap all kwargs necessary for the simulation
        kwargs = cf.get_kwargs(cf.internal_names,cf.internal_parameter_list[ip])
        kwargs.update(job_kwargs)
        kwargs.update(cf.std_kwargs)

        #add seed
        if hasattr(cf,'seed') and cf.seed is not None and cf.seed>=0:
            key = 'seed'
            if key in kwargs:
                key = 'randomseed'
            kwargs[key] = N_int_param*j + ip + cf.seed
        
        t_start = time.time()

        try:
            # start the simulation and save the result
            results[ip] = simcode.simulation_code(kwargs)
        except Exception as e:
            # in case there's an error, write that into the progress file
            if not is_local:
                with open(_get_progress_fn(cf,j),"w") as progressfile:
                    progressfile.write("error: "+e.__class__.__name__+"\n")

            traceback.print_exc(file=sys.stderr)
            sys.exit(1) 

        t_end = time.time()
        
        times[ip] = t_end - t_start

        if is_local:
            if ip>0:
                _update_progress((ip + 1.)/N_int_param,40,"est. time per measurement: "+ str(mean(times[:ip+1])) + " +/- " + str(std(times[:ip+1])/sqrt(ip+1)) )
            else:
                _update_progress((ip + 1.)/N_int_param,40,"est. time per measurement: "+ str(mean(times[:ip+1])) )
        else:
            _update_progress_file(ip,N_int_param,times[:ip+1],_get_progress_fn(cf,j))
        
    #save results
    if not os.path.exists(cf.resultpath):
        os.mkdir(cf.resultpath)

    if any([result is not None for result in results]) and not cf.only_save_times:
        with open(cf.resultpath+'/results_%d.p' % (j),'wb') as dumpfile:
            pickle.dump(results,dumpfile,protocol=pickle.HIGHEST_PROTOCOL)

    #save times needed
    with open(cf.resultpath+'/times_%d.p' % (j),'wb') as dumpfile:
        pickle.dump(times,dumpfile,protocol=pickle.HIGHEST_PROTOCOL)


if __name__=="__main__":

    #get job number and alternatively another resultpath
    args = sys.argv[1:]
    j = int(args[0]) #jobnumber
    if len(args)>1:
        resultpath = args[1] 
    else:
        resultpath = None

    job(j,resultpath)



import os
import sys
import qsuite
from qsuite import sftp_put_files
from qsuite import ssh_command

def get_dummy(qname):
    potential_file = os.path.join(qsuite.customdir,qname+"dummy.sh")
    if not os.path.exists(potential_file):
        potential_file = os.path.join(qsuite.__path__[0],"jobscripts",qname+"dummy.sh")
    s = open(potential_file,'r').read()

    return s

def get_jobscript(cf,array_id=None):
    JOBSCRIPT = get_dummy(cf.queue)

    if array_id is None:
        arr_id_min = cf.jmin+1
        arr_id_max = cf.jmax+1
    elif type(array_id) in [ list, tuple ] and len(array_id)==2 :
        arr_id_min = array_id[0]
        arr_id_max = array_id[1]
    else:
        arr_id_min = array_id
        arr_id_max = array_id

    jobscript = JOBSCRIPT % (\
                cf.shell,
                cf.memory,
                arr_id_min,
                arr_id_max,
                cf.serverpath+"/output",
                cf.serverpath+"/output",
                cf.priority,
                cf.pythonpath,
                cf.serverpath,
                )

    return jobscript

def get_file_list(cf):
    
    files_destinations = [            
            ( 
                os.path.join(qsuite.__path__[0],"queuesys","wrap_results.py"),
                cf.serverpath + "/wrap_results.py" 
            ),
            (
                os.path.join(qsuite.__path__[0],"queuesys","job.py"),
                cf.serverpath + "/job.py"
            ),
            (
                os.path.join(qsuite.__path__[0],"qconfig.py"),
                cf.serverpath + "/qconfig.py"
            ),
        ]

    local_files = [  ( os.path.join(os.getcwd(),lname),  cf.serverpath + "/" + sname ) for sname,lname in cf.files_to_scp.items() ]

    add_files = [ ( os.path.join(os.getcwd(),aname),  cf.serverpath + "/" + aname ) for aname in cf.additional_files_to_scp  ]

    files_destinations += local_files + add_files

    return files_destinations

def make_job_ready(cf,ssh,array_id=None):

    files_destinations = get_file_list(cf)
    joblocal_names = []

    if type(array_id) is not list and type(array_id) is not tuple:
        is_list = True
        array_id = [ array_id ]
    else:
        is_list = False

    for a_id in array_id:
        jobscript = get_jobscript(cf,a_id)
        print("\nUsing jobscript:\n================")    
        print(jobscript)

        if a_id is not None and (type(a_id) not in [ tuple, list ]):
            joblocalname = os.path.join(os.getcwd(),"."+cf.basename+"_%d.sh" % (a_id))
            jobservername = cf.serverpath+"/"+cf.basename+"_%d.sh" %(a_id)
        elif (type(a_id) in [ tuple, list ]):
            joblocalname = os.path.join(os.getcwd(),"."+cf.basename+"_%d_%d.sh" % tuple(a_id))
            jobservername = cf.serverpath+"/"+cf.basename+"_%d_%d.sh" % tuple(a_id)
        else:
            joblocalname = os.path.join(os.getcwd(),"."+cf.basename+".sh")
            jobservername = cf.serverpath+"/"+cf.basename+".sh"

        f = open(joblocalname,'w')
        f.write(jobscript)
        f.close()

        files_destinations += [ (joblocalname, jobservername) ]
        joblocal_names.append(joblocalname)

    sftp_put_files(ssh,cf,files_destinations)
    
    for joblocalname in joblocal_names:
        qsuite.rm(joblocalname)

    files_chmod_x = [ "chmod +x "+f[1]+"; " for f in files_destinations if f[1].endswith(".sh") ]

    ssh_command(ssh,''.join(files_chmod_x))

    if "execute_after_scp.sh" in cf.files_to_scp:
        ssh_command(ssh,"cd " +cf.serverpath+"; ./execute_after_scp.sh;")


def start_job(cf,ssh,array_id=None):
    """job id procedure taken from https://github.com/osg-bosco/BLAH/blob/1d217fad9c6b54a5e543f7a9d050e77047be0bb1/src/scripts/pbs_submit.sh#L193"""
    if type(array_id) is list or type(array_id) is tuple:
        suffices = ["_%d.sh" % a_id if type(a_id) in [ int ] else "_%d_%d.sh" % tuple(a_id) for a_id in array_id]
    else:
        suffices = [".sh"]
        
    for suffix in suffices:
        if cf.queue=="SGE":
            ssh_command(ssh,"cd " +cf.serverpath+";\
                             jobID=`qsub " + cf.basename + suffix +"`;\
                             jobID=`echo $jobID | awk 'match($0,/[0-9]+/){print substr($0, RSTART, RLENGTH)}'`;\
                             echo $jobID > .jobid;")
        elif cf.queue=="PBS":
            ssh_command(ssh,"cd " +cf.serverpath+";\
                             jobID=`qsub " + cf.basename + suffix +"`;\
                             echo $jobID > .jobid;")
        else:
            print("Unknown queue:",cf.queue)
            sys.exit(1)

    N = len(cf.parameter_list)-1
    filepath = cf.serverpath + "/output/progress_"
    if array_id is None:
        cmd = ('for i in `seq -w 0 %d`; do echo " " > '+filepath+'${i}_%d; done;') % (N,N)
    elif type(array_id) is list or type(array_id) is tuple:
        cmd = ""
        for a_id in array_id:
            if type(a_id) in [ tuple, list ]:
                cmd += (' for i in `seq -w %d %d`; do echo " " > '+filepath+'${i}_%d; done;') % tuple(list(a_id) + [N])
            else:
                cmd += (' echo " " > '+filepath+'%d_%d; done;') % (int(a_id)-1,N)
    else:
        cmd = ('echo " " > '+filepath+'%d_%d; done;') % (int(array_id)-1,N)
    ssh_command(ssh,cmd)
    



from __future__ import print_function
from tabulate import tabulate
import copy
from qsuite import ssh_command

import tempfile

def _get_progress_fn(cf, j):
    str_len = str(len(str(len(cf.parameter_list)-1)))
    fmt = "%0"+str_len+"d_%d"
    return cf.serverpath + ("/output/progress_"+fmt) % (j, len(cf.parameter_list)-1)


def print_params(cf):

    #for each parameter combination, add job id (j) in the beginning and arrayjob id in the end
    params = [ [j] + list(p) + [j+1] for j,p in enumerate(cf.parameter_list) ]

    #get names from config, besides if the entry is None and the length is equal to the number of measurements, then add MeasID
    names = [ n if not (n is None and len(cf.external_parameters[i][1])==cf.N_measurements) else "Meas.ID" for i,n in enumerate(cf.parameter_names)  ]

    print(tabulate(params, headers=["Job ID"] + names + ["Array ID"]))


def _get_progress_sftp(cf,ssh):

    N = len(cf.parameter_list)-1
    filepath = cf.serverpath + "/output/progress_*_"+str(N)
    cmd = "cd " +cf.serverpath + "/output/; cat progress_*_"+str(N)+" > all_progress" 
    ssh_command(ssh,cmd,noprint=True)

    ftp = ssh.open_sftp()

    with tempfile.TemporaryFile() as tmp:

        ftp.getfo(cf.serverpath + "/output/all_progress", tmp)
        tmp.seek(0)

        progresses_ = [ s.decode('utf-8')[:-1] for s in tmp.readlines() ]
        progresses = []
        for p in progresses_:
            if len(p)>1:
                text_and_time = p.split("__")
                if len(text_and_time)==1:
                    text_and_time.append("")
            else:
                text_and_time = ['waiting...','']

            progresses.append(text_and_time)

    return progresses

def print_status(cf,ssh):

    #for each parameter combination, add job id (j) in the beginning and arrayjob id in the end
    progresses = _get_progress_sftp(cf,ssh)
    prog = []

    record_waiting_id = False
    record_done_id = False

    # go through the progresses and clump together all 'waitings' and 'dones'
    j = 0
    while j<len(progresses):

        # while there's normal entries, count up and add them to output
        while j<len(progresses) and progresses[j][0]!='waiting...' and progresses[j][-1]!="done":
            prog.append( [j+1] + progresses[j] )
            j += 1

        # start clumping 'waitings'
        if j<len(progresses) and progresses[j][0]=='waiting...':

            old_id = j+1

            while j<len(progresses) and progresses[j][0]=='waiting...':
                j += 1

            if old_id==j:
                prog.append( [ old_id ] + progresses[old_id-1])
                #if j<len(progresses):
                #    prog.append( [ j+1 ] + progresses[j])
            else:
                prog.append( [ str(old_id)+"-"+str(j) ] + progresses[j-1])

        # start clumping 'dones'
        if j<len(progresses) and progresses[j][-1]=='done':

            old_id = j+1

            while j<len(progresses) and progresses[j][-1]=='done':
                j += 1

            if old_id==j:
                prog.append( [ old_id ] + progresses[old_id-1])
                #if j<len(progresses):
                #    prog.append( [ j+1 ] + progresses[j])
            else:
                prog.append( [ str(old_id)+"-"+str(j) ] + progresses[j-1])


    #prog = [ [j+1]+p for j,p in enumerate(progresses) ]

    names = [ "Array ID", "Progress", "Rem. Time" ]

    print(tabulate(prog, headers=names))


def print_params_and_status(cf,ssh):
    
    #for each parameter combination, add job id (j) in the beginning and arrayjob id in the end
    progresses = _get_progress_sftp(cf,ssh)
    table = [ [j] + list(p[0]) + [j+1] + list(p[1]) for j,p in enumerate(zip(cf.parameter_list,progresses)) ]

    #get names from config, besides if the entry is None and the length is equal to the number of measurements, then add MeasID
    names = [ n if not (n is None and len(cf.external_parameters[i][1])==cf.N_measurements) else "Meas.ID" for i,n in enumerate(cf.parameter_names)  ]
    names = ["Job ID"] + names + ["Array ID",  "Progress", "Rem. Time" ]

    print(tabulate(table, headers = names))



from setuptools import setup

setup(name='qsuite',
      version='0.4.14',
      description='A CLI to submit and retrieve large-scale jobs/results to and from HPC clusters.',
      url='https://github.com/benmaier/qsuite',
      author='Benjamin F. Maier',
      author_email='benjaminfrankmaier@gmail.com',
      license='MIT',
      packages=['qsuite'],
      include_package_data = True,
      install_requires=[
        'numpy>=1.17',
        'paramiko>=2.7.2',
        'pathos>=0.2.7',
        'tabulate>=0.8.7',
      ],
      dependency_links=[
          ],
      entry_points = {
            'console_scripts': [
                    'qsuite = qsuite.qsuite_binary:main',
                ],
          },
      zip_safe=False)



from __future__ import print_function
import os 
import sys
import numpy as np
import ast
import itertools

try:
    # Python 2
    from itertools import izip
except ImportError:
    # Python 3
    izip = zip

class qconfig(object):

    def __init__(self,filename="qsuite_config.py",qsuiteparser=None):

        if qsuiteparser is None:
            self.configpath = os.path.join(os.getcwd(), filename)
        else:
            self.configfile = qsuiteparser.get('Files','config')
            if not os.path.exists(self.configfile):
                self.configfile = os.path.join(os.getcwd(),self.configfile)

            self.files_to_scp = { 
                                    "qsuite_config.py": self.configfile,
                                    "simulation.py": qsuiteparser.get('Files','simulation'),
                                }

            if not qsuiteparser.get('Files','execute_after_scp') == "None": 
                self.files_to_scp["execute_after_scp.sh"] = qsuiteparser.get('Files','execute_after_scp')

            if not qsuiteparser.get('Files','customwrap') == "None":
                self.files_to_scp["custom_wrap_results.py"] = qsuiteparser.get('Files','customwrap')

            add_files = ast.literal_eval(qsuiteparser.get('Files','additional_files'))
            self.additional_files_to_scp = [ f for f in add_files if not f=='' ]

            self.configpath = self.configfile



        if os.path.exists(self.configpath):
            cf = self.get_cf(self.configpath)
        else:
            print("No "+filename+" found in current working directory!")
            sys.exit(1)


        #inherit all properties from the config module
        self.inherit_properties(cf)

        #prepare parameters lists
        self.parameter_names, self.parameter_list = self.get_param_lists(self.external_parameters)
        self.internal_names, self.internal_parameter_list = self.get_param_lists(self.internal_parameters)

        #prepare standard parameters list and kwargs
        self.standard_names, self.standard_list = self.get_param_lists(self.standard_parameters,product=False)
        self.std_kwargs = self.get_kwargs(self.standard_names,self.standard_list)

        #get min and max jobnumber
        self.jmin = 0
        self.jmax = len(self.parameter_list)-1

    #======================== INHERIT PROPERTIES FROM THE CONFIG FILE ============================

    def get_cf(self,path):

        if sys.version_info[0] == 2:
            import imp
            cf = imp.load_source("queueconfig",path)
        elif sys.version_info >= (3,5):
            import importlib.util
            specifications = importlib.util.spec_from_file_location("queueconfig",path)
            cf = importlib.util.module_from_spec(specifications)
            specifications.loader.exec_module(cf)
        else:
            print("Python version",sys.version_info[0],"not supported")
            sys.exit(1)

        return cf

    def inherit_properties(self,cf):
        new_props = [ prop for prop in dir(cf) if not prop.startswith('__') ]

        for p in new_props:
            self.__dict__[p] = cf.__dict__[p]

    #============================= PARAMETER STRING CONVERSION FUNCTIONS ==========================

    """
    def get_param_lists_old(self,pliststr,plist=None):
        if pliststr[-1]=="]" and pliststr[0]=="[":
            try:
                pstring = pliststr[1:-1] #delete first and last character
                array_names = re.sub("[\(\[].*?[\)\]]", "", pstring) #delete array delimiters and what's in between them
                array_names = re.sub(" ", "", array_names).split(',') #delete spaces and get list
                if plist is not None:
                    parameter_list = list(itertools.product(*plist))
            except:
                print "wrong format in parameter list:", pliststr
                sys.exit(1)
        else:
            print "wrong format in outer scope of parameter list:", pliststr
            print "forgot the mandatory brackets around the arrays: [] ?)"
            sys.exit(1)

        if plist is not None:
            return array_names, parameter_list
        else:
            return array_names

    def get_kwargs_old(self,plist,pnames,name_to_option):
        if not (len(pnames)==1 and pnames[0]==""):
            kwargs = { 
                       name_to_option[name]:plist[iname]\
                       for iname,name in itertools.izip(range(len(pnames)),pnames)\
                       if name_to_option[name] is not None 
                     }
        else:
            kwargs = {}

        return kwargs
    """

    def get_param_lists(self,plist,product=True):
        pnames = [ p[0] for p in plist ]
        plist = [ p[1] for p in plist ]

        if product:
            parameter_list = list(itertools.product(*plist))
        else:
            parameter_list = plist

        return pnames, parameter_list

    def get_kwargs(self,pnames,current_parameters):
        kwargs = { 
                   name : current_parameters[iname]\
                   for iname,name in izip(range(len(pnames)),pnames)\
                   if name is not None 
                 }

        return kwargs

    def get_number_of_jobs(self):
        return len(self.parameter_list)

    
if __name__=="__main__":

    cf = qconfig()

    print(cf.get_param_lists(cf.parameter_names,cf.parameter_list[0]))
    



from numpy import *
import os
from qconfig import qconfig
#import gzip

#============== for python 2/3 compatibility =====================
try:
    # Python 2
    import cPickle as pickle
except ImportError:
    # Python 3
    import pickle

try:
    basestring
except NameError:
    basestring = str

cf = qconfig()

#=============================================================================

def mean_and_err(arr,axis=None):

    if isinstance(axis,basestring) or axis is None:
        axis = (cf.parameter_names+cf.internal_names).index(axis)
    else:
        axis = axis

    N = len((cf.external_parameters+cf.internal_parameters)[axis][1])

    return arr.mean(axis=axis), arr.std(axis=axis)/sqrt(N)

#====== definition for the treatment of the results goes here ================= 

def prepare_results(res):
    return mean_and_err(res)

#==============================================================================

times = pickle.load(open(cf.resultpath+"/times.p",'rb'))
time,err = mean_and_err(array(times))
savez(cf.resultpath+"/mean_err_times.npz",time,err)

if not cf.only_save_times:
    results = pickle.load(open(cf.resultpath+"/results.p",'rb'))
    results = prepare_results(array(results))
    if isinstance(results,(list,tuple)):
        savez(cf.resultpath+"/mean_err_result_list.npz",*results)
    else:
        save(cf.resultpath+"/mean_err_results.npy",results)


